def robodyno():
    import argparse
    import colorama
    from .list_devices import list_devices
    from .monitor import monitor
    from ..interfaces import CanBus

    colorama.init(autoreset=True)

    parser = argparse.ArgumentParser(description='Robodyno command line tools.')
    parser.add_argument('command', choices=['list', 'monitor', 'motor'], help='''
                            list all robodyno devices
                            / monitor messages on bus
                            / motor basic operations
                        ''')
    parser.add_argument('--id', type=str, nargs='*', default=[], 
                        help='robodyno device id filter')
    parser.add_argument('-c', '--channel', type=str, default='can0',
                        help='can bus channel')
    parser.add_argument('-b', '--bitrate', type=str, default='CAN_1M',
                        help='can bus bitrate')

    args = parser.parse_args()

    def parse_id(str_list):
        """parse list of hex id from string format
        
        Args:
            str_list: string id list
        
        Returns:
            hex id list
        """
        if len(str_list) > 0 and str_list[0] == 'all':
            return [id for id in range(0x01, 0x40)]
        else:
            return list(map(lambda id_str: int(id_str, 0), str_list))

    try:
        can = CanBus(bitrate = args.bitrate, channel = args.channel)
    except:
        raise ValueError('Wrong can bus configurations.')

    try:
        id_list = parse_id(args.id)
    except:
        raise ValueError('Wrong id list.')

    if args.command == 'list':
        list_devices(can, id_list)

    if args.command == 'monitor':
        monitor(can, id_list)

def robodyno_motor():
    import time
    import argparse
    import colorama
    from termcolor import colored
    from ..interfaces import CanBus
    from ..components import Motor

    colorama.init(autoreset=True)

    parser = argparse.ArgumentParser(description='Robodyno command line tools.')
    parser.add_argument('command', 
                        choices=['enable', 'disable', 'info', 'control', 'config', 'reset'], 
                        help='Operate robodyno motor via command line.')
    parser.add_argument('--id', type=str, nargs='*', default=['0x10'],
                        help='robodyno device id filter, dafault [0x10]')
    parser.add_argument('-c', '--channel', type=str, default='can0',
                        help='can bus channel')
    parser.add_argument('-b', '--bitrate', type=str, default='CAN_1M',
                        help='can bus bitrate')

    parser.add_argument('-p', '--pos', type=float,
                        help='motor position setpoint(rad)')
    parser.add_argument('-v', '--vel', type=float,
                        help='motor velocity setpoint(rad/s)')
    parser.add_argument('-t', '--torque', type=float,
                        help='motor torque setpoint(Nm)')
    parser.add_argument('--new-id', type=str,
                        help='change motor id with new id')
    parser.add_argument('-s', '--save', action='store_true',
                        help='save motor configuration when set new id')

    args = parser.parse_args()

    try:
        can = CanBus(bitrate = args.bitrate, channel = args.channel)
    except:
        raise ValueError('Wrong can bus configurations.')

    def id_to_motor_list(str_list):
        """Parse id list to motor object list.
        
        Args:
            str_list: string id list
        
        Returns:
            robodyno motor object list
        """
        id_list = []
        motor_list = []
        if len(str_list) > 0 and ('all' in str_list):
            id_list = [id for id in range(0x01, 0x40)]
        else:
            id_list = list(map(lambda id_str: int(id_str, 0), str_list))
        for id in id_list:
            try:
                motor_list.append(Motor(can, id))
            except:
                pass
        return motor_list

    try:
        motor_list = id_to_motor_list(args.id)
    except:
        raise ValueError('Wrong id list.')

    if len(motor_list) == 0:
        raise ValueError('No available motor on can bus.')

    if args.command == 'enable':
        for motor in motor_list:
            motor.enable()

    if args.command == 'disable':
        for motor in motor_list:
            motor.disable()

    if args.command == 'info':
        for motor in motor_list:
            time.sleep(1)
            ver = motor.get_version(0.1)
            vbus = motor.get_voltage(0.1)
            temp = motor.get_temperature(0.1)
            pos, vel, torque = motor.get_feedback(0.1)

            name_str = colored('[0x{:02X}]'.format(motor.id), 'cyan') + ' ' + colored(ver['type'].name, 'green')
            ver_str = 'ver   : ' + colored('{}.{}'.format(ver['main_version'], ver['sub_version']), 'yellow')
            print('{:<60}{}'.format(name_str, ver_str))

            mode_str = 'mode: ' + colored(motor.mode.name, 'yellow')
            state_str = 'state : '
            if motor.error['error'].value:
                state_str += colored(motor.error['error'].name, 'red')
            else:
                state_str += colored(motor.state.name, 'yellow')
            print('{:<51}{}'.format(mode_str, state_str))
            
            vbus_str = 'vbus: ' + colored('{:.2f}V'.format(vbus), 'yellow')
            temp_str = 'temp: ' + colored('{:.2f}°C'.format(temp), 'yellow')
            curr_str = 'curr  : ' + colored('{:.2f}A'.format(torque / motor.torque_constant), 'yellow')
            print('{:<30}{:<30}{}'.format(vbus_str, temp_str, curr_str))

            pos_str = 'pos : ' + colored('{:.4f}rad'.format(pos), 'yellow')
            vel_str = 'vel : ' + colored('{:.4f}rad/s'.format(vel), 'yellow')
            torque_str = 'torque: ' + colored('{:.4f}Nm'.format(torque), 'yellow')
            print('{:<30}{:<30}{}'.format(pos_str, vel_str, torque_str))

    if args.command == 'control':
        for motor in motor_list:
            if args.pos is not None:
                vel_ff = args.vel if args.vel is not None else 0
                torque_ff = args.torque if args.torque is not None else 0
                for motor in motor_list:
                    motor.position_filter_mode(5)
                    motor.enable()
                    motor.set_pos(args.pos, vel_ff, torque_ff)
            elif args.vel is not None:
                torque_ff = args.torque if args.torque is not None else 0
                for motor in motor_list:
                    motor.velocity_ramp_mode(5)
                    motor.enable()
                    motor.set_vel(args.vel, torque_ff)
            elif args.torque is not None:
                for motor in motor_list:
                    motor.torque_mode(5)
                    motor.enable()
                    motor.set_torque(args.torque)

    if args.command == 'config':
        if len(motor_list) != 1:
            raise ValueError('Can not set multiple motor ids at the same time.')
        motor = motor_list[0]
        try:
            new_id = int(args.new_id, 0)
            if new_id < 0x01 or new_id >= 0x40:
                raise ValueError()
        except:
            raise ValueError('Wrong motor new id, choose from [0x01,0x40).')
        motor.config_can_bus(new_id, bitrate = args.bitrate)
        time.sleep(0.5)
        motor = Motor(can, new_id)
        if args.save:
            motor.save_configuration()

    if args.command == 'reset':
        if len(motor_list) != 1:
            raise ValueError('Can not reset multiple motor at the same time.')
        motor = motor_list[0]
        print('resetting...')
        motor.reset()
        time.sleep(3)
        motor = Motor(can, 0x10)
        print('calibrating...')
        motor.calibrate()
        time.sleep(1)
        while motor.state.value != 1:
            pass
        time.sleep(0.2)
        motor.save_configuration()
        time.sleep(0.5)
        print('finished.')



#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""can_bus_interface.py
Time    :   2022/10/08
Author  :   song 
Version :   1.0
Contact :   zhaosongy@126.com
License :   (C)Copyright 2022, robottime / robodyno

Can Bus Interface Abstract Class
"""

from abc import abstractmethod
import threading
import time
from collections import defaultdict
from ..interface import Interface
from .motor_node import CanBusMotorNode
from .device_node import CanBusDeviceNode

class CanBusInterface(Interface):
    """Can bus interface manage can id, rx thread and subscribers."""
    
    MAX_DEVICE_NUM = 0x40
    DEVICE_ID_ALL = MAX_DEVICE_NUM

    MAX_COMMAND_NUM = 0x20
    COMMAND_ID_ALL = MAX_COMMAND_NUM

    def __init__(self, auto_connect = True, *args, **kwargs):
        """Init can driver interface.
        
        Args:
            auto_connect: if set to False, you need to call connect() manually
        """
        self._subscribers = defaultdict(lambda: defaultdict(list))
        self._send_lock = threading.Lock()
        self._thread_running = False
        self._rx_thread = threading.Thread(
            target = self._rx_thread_method,
            name = 'candle.rx'
        )
        self._rx_thread.daemon = True
        if auto_connect:
            self.connect()
            time.sleep(0.1)

    def pack_can_id(self, device_id, command_id):
        """calculate can frame id from device id and command id
        
        Args:
            device_id: robodyno device id
            command_id: device command id
        
        Returns:
            can frame id
        """
        return device_id << 5 | command_id
    
    def unpack_can_id(self, can_id):
        """split can frame id into device id and command id
        
        Args:
            can_id: can frame id
        
        Returns:
            (device_id, command_id)
        """
        return ((can_id & 0x7ff) >> 5, can_id & 0x1f)
    
    def get_motor_node(self, motor):
        """Return new can bus motor node."""
        return CanBusMotorNode(motor = motor, iface = self)
    
    def get_device_node(self, device):
        """Return new can bus motor node."""
        return CanBusDeviceNode(device = device, iface = self)
    
    @abstractmethod
    def _enable(self):
        """Enable can bus interface."""

    @abstractmethod
    def _disable(self):
        """Disable can bus interface."""
        
        
    def connect(self):
        """Connect to can bus."""
        try:
            self._enable()
            self._thread_running = True
            self._rx_thread.start()
        except:
            raise IOError('Failed to connect can bus.')
    
    def disconnect(self):
        """Disconnect to can bus."""
        self._thread_running = False
        self._rx_thread.join(1)
        try:
            self._disable()
        except:
            raise IOError('Failed to disconnect can bus.')
    
    @abstractmethod
    def _send(self, can_frame_id, data, remote):
        """Send can frame with id, data and remote bit.
        
        Args:
            can_frame_id: standard can frame id(11bit)
            data: can frame payload byte array, up to 8 bytes
            remote: can frame RTR bit
        """

    @abstractmethod
    def _read(self, timeout = 0):
        """Read can frame with timeout.
        
        Args:
            timeout: blocking read timeout(s)
        
        Returns:
            can_frame_id, data, timestamp
        """

    def send_can_msg(self, device_id, command_id, data = b'', remote = False):
        """Send message from can bus.
        
        Args:
            device_id: robodyno device id, upper 6 bits of can frame id
            command_id: device command id, lower 5 bits of can frame id
            data: bytes of can frame payload
            remotr: can frame rtr bit
        """
        with self._send_lock:
            can_frame_id = self.pack_can_id(device_id, command_id)
            self._send(can_frame_id, data, remote)

    def subscribe(self, device_id, command_id, callback):
        """Subscribe specific device id and command id from can bus.
        
        Args:
            device_id: robodyno device id or Interface.DEVICE_ID_ALL
            command_id: device command id or Interface.COMMAND_ID_ALL
            callback: func(device_id, command_id, data, timestamp)
        """
        if callback not in self._subscribers[device_id][command_id]:
            self._subscribers[device_id][command_id].append(callback)
        
    def unsubscribe(self, device_id, command_id, callback = None):
        """Unsubscribe specific device id and command id from can bus.
        
        Args:
            device_id: robodyno device id or Interface.DEVICE_ID_ALL
            command_id: device command id or Interface.COMMAND_ID_ALL
            callback: callback to unsubscribe or leave None to unsubscribe all callbacks
        """
        if callback:
            try:
                self._subscribers[device_id][command_id].remove(callback)
            except:
                raise ValueError('Failed to find callback to unsubscribe.')
        else:
            self._subscribers[device_id][command_id] = []
    
    def _notify(self, frame_id, data, timestamp):
        """Protected method, notify subscribers when can frame received.
        
        Args:
            frame_id: can frame id
            data: can frame payload
            timestamp: can frame timestamp
        """
        device_id, command_id = self.unpack_can_id(frame_id)
        for subscriber in self._subscribers[device_id][command_id]:
            subscriber(device_id, command_id, data, timestamp)
        for subscriber in self._subscribers[device_id][self.COMMAND_ID_ALL]:
            subscriber(device_id, command_id, data, timestamp)
        for subscriber in self._subscribers[self.DEVICE_ID_ALL][command_id]:
            subscriber(device_id, command_id, data, timestamp)
        for subscriber in self._subscribers[self.DEVICE_ID_ALL][self.COMMAND_ID_ALL]:
            subscriber(device_id, command_id, data, timestamp)
        
    def _rx_thread_method(self):
        """read can bus thread method"""
        while self._thread_running:
            try:
                can_frame_id, data, timestamp = self._read(0.1)
                self._notify(can_frame_id, data, timestamp)
            except:
                pass



#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""motor_node.py
Time    :   2022/09/30
Author  :   song 
Version :   1.0
Contact :   zhaosongy@126.com
License :   (C)Copyright 2022, robottime / robodyno
"""

from math import pi, fabs
import struct
from ...components.motor import *
from .utils import CanSpeed, get_from_bus, send_to_bus

class CanBusMotorNode(MotorNode):
    """Can Bus Motor Node"""

    CMD_GET_VERSION = 0x01
    CMD_HEARTBEAT = 0x02
    CMD_ESTOP = 0x03
    CMD_REBOOT = 0x04
    CMD_CLEAR_ERRORS = 0x05
    CMD_SAVE = 0x06
    CMD_CONFIG_CAN = 0x07
    CMD_SET_STATE = 0x08
    CMD_GET_HARDWARE_STATUS = 0x09
    CMD_GET_MOTOR_FEEDBACK = 0x0A
    CMD_GET_MODE = 0x0B
    CMD_SET_MODE = 0x0C
    CMD_GET_PID = 0x0D
    CMD_SET_PID = 0x0E
    CMD_GET_LIMITS = 0x0F
    CMD_SET_LIMITS = 0x10
    CMD_SET_POS = 0x11
    CMD_SET_VEL = 0x12
    CMD_SET_TORQUE = 0x13
    CMD_RESET = 0x14
    CMD_UNLOCK = 0x15
    
    def __init__(self, motor, iface):
        """Init node from motor object and can bus interface.
        
        Args:
            motor: robodyno motor object
            iface: can bus interface
        """
        super().__init__(motor, iface)
        self.id = motor.id
        self._reduction = motor.reduction
        self._rot_factor = motor.reduction / 2 / pi
        self._iface().subscribe(
            device_id = self.id, 
            command_id = self.CMD_HEARTBEAT, 
            callback = self._heartbeat_callback
        )

    def _delete(self):
        """Collect node from memory."""
        self._iface().unsubscribe(
            device_id = self.id, 
            command_id = self.CMD_HEARTBEAT, 
            callback = self._heartbeat_callback
        )

    def _heartbeat_callback(self, device_id, command_id, data, timestamp):
        """Update motor attributes from heartbeat."""
        state, err, merr, eerr, cerr, cm, im, _ = struct.unpack('<BBBBBBBB', data)
        self._motor().state = MotorState(state)
        self._motor().error.update({
            'error': MotorError(err),
            'motor_err': MotorMotorError(merr),
            'encoder_err': MotorEncoderError(eerr),
            'controller_err': MotorControllerError(cerr),
        })
        self._motor().mode = self._ctrl_mode_from_raw(cm, im)
        self._motor().sync_time = timestamp

    def _ctrl_mode_to_raw(self, mode):
        """Translate from control mode enum to control mode raw msg
        
        Args:
            mode: control mode enum
        
        Returns:
            tuple of raw msg: (control mode, input mode)
            None when mode not recognized
        """
        if mode == MotorControlMode.POSITION_MODE:
            return (3, 1)
        elif mode == MotorControlMode.POSITION_FILTER_MODE:
            return (3, 3)
        elif mode == MotorControlMode.POSITION_TRACK_MODE:
            return (3, 5)
        elif mode == MotorControlMode.VELOCITY_MODE:
            return (2, 1)
        elif mode == MotorControlMode.VELOCITY_RAMP_MODE:
            return (2, 2)
        elif mode == MotorControlMode.TORQUE_MODE:
            return (1, 1)
        else:
            return None

    def _ctrl_mode_from_raw(self, cmode, imode):
        """Translate from control mode raw msg to control mode enum
        
        Args:
            cmode: raw control mode
            imode: raw input mode
        
        Returns:
            control mode enum
        """
        if cmode == 3:
            if imode == 1:
                return MotorControlMode.POSITION_MODE
            elif imode == 3:
                return MotorControlMode.POSITION_FILTER_MODE
            elif imode == 5:
                return MotorControlMode.POSITION_TRACK_MODE
        if cmode == 2:
            if imode == 1:
                return MotorControlMode.VELOCITY_MODE
            elif imode == 2:
                return MotorControlMode.VELOCITY_RAMP_MODE
        if cmode == 1 and imode == 1:
            return MotorControlMode.TORQUE_MODE
        return MotorControlMode.UNKNOWN
    
    @get_from_bus(CMD_GET_VERSION, '<HHI')
    def get_version(self, main_ver, sub_ver, type):
        """Get motor firmware / simulation version.
        
        Args:
            timeout: 0 indicates unlimited timeout(s)
        
        Returns:
            dictionary of motor version
            None if timeout
        """
        return {
            'main_version': main_ver,
            'sub_version': sub_ver,
            'type': MotorType(type)
        }
    
    @send_to_bus(CMD_ESTOP)
    def estop(self):
        """Emergency stop motor."""
        pass

    @send_to_bus(CMD_REBOOT)
    def reboot(self):
        """Reboot motor."""
        pass

    @send_to_bus(CMD_CLEAR_ERRORS)
    def clear_errors(self):
        """Try to clear motor errors."""
        pass

    @send_to_bus(CMD_SAVE)
    def save_configuration(self):
        """Save configurations to hardware."""
        pass

    @send_to_bus(CMD_CONFIG_CAN, '<HHI')
    def config_can_bus(self, new_id, heartbeat, bitrate):
        """Configure motor can bus settings.
        
        Args:
            new_id: motor device id(0x01-0x3f)
            heartbeat: heartbeat period(ss)
            bitrate: choose from 'CAN_1000K', 'CAN_500K', 'CAN_250K',
                     (save and reboot to take effect)
        """
        bitrate = CanSpeed[bitrate].value
        if bitrate == 250000:
            bitrate_id = 0
        elif bitrate == 500000:
            bitrate_id = 1
        else:
            bitrate_id = 2
        return (new_id, bitrate_id, int(heartbeat*1000))

    @send_to_bus(CMD_SET_STATE, '<I')
    def set_state(self, state):
        """Cange motor state.
        
        Args:
            state: motor state(MotorState)
        """
        return (state.value,)

    @get_from_bus(CMD_GET_HARDWARE_STATUS, '<ff')
    def get_voltage(self, vbus, temperature):
        """Get motor vbus voltage.
        
        Args:
            timeout: 0 indicates unlimited timeout(s)
        
        Returns:
            voltage(V)
            None if timeout
        """
        return vbus
    
    @get_from_bus(CMD_GET_HARDWARE_STATUS, '<ff')
    def get_temperature(self, vbus, temperature):
        """Get motor temperature.
        
        Args:
            timeout: 0 indicates unlimited timeout(s)
        
        Returns:
            temperature(°C)
            None if timeout
        """
        return temperature
    
    @get_from_bus(CMD_GET_MOTOR_FEEDBACK, '<fee')
    def get_feedback(self, pos, vel, torque):
        """Get motor position, velocity and torque feedback.
        
        Args:
            timeout: 0 indicates unlimited timeout(s)
        
        Returns:
            position(rad), velocity(rad/s), torque(Nm)
            None if timeout
        """
        return pos / self._rot_factor, vel / self._rot_factor, torque * self._reduction
    
    @get_from_bus(CMD_GET_MODE, '8s')
    def get_mode(self, payload):
        """Get motor control mode and params.
        
        Args:
            timeout: 0 indicates unlimited timeout(s)
        
        Returns:
            control mode(MotorControlMode), control params dict
            None if timeout
        """
        control_mode, input_mode = struct.unpack('<BB', payload[:2])
        mode = self._ctrl_mode_from_raw(control_mode, input_mode)
        self._motor().mode = mode
        if mode == MotorControlMode.POSITION_FILTER_MODE:
            bandwidth, = struct.unpack('<f', payload[2:6])
            return (mode, {'bandwidth': bandwidth})
        elif mode == MotorControlMode.POSITION_TRACK_MODE:
            vel, acc, dec = struct.unpack('<eee', payload[2:8])
            return (mode, {
                'vel': fabs(vel / self._rot_factor),
                'acc': fabs(acc / self._rot_factor),
                'dec': fabs(dec / self._rot_factor),
            })
        elif mode == MotorControlMode.VELOCITY_RAMP_MODE:
            ramp, = struct.unpack('<f', payload[2:6])
            return (mode, {'ramp': fabs(ramp / self._rot_factor)})
        else:
            return (mode, )
    
    @send_to_bus(CMD_SET_MODE, '<BB')
    def position_mode(self):
        """Enter position pid mode."""
        return self._ctrl_mode_to_raw(MotorControlMode.POSITION_MODE)

    @send_to_bus(CMD_SET_MODE, '<BBf')
    def position_filter_mode(self, bandwidth):
        """Enter position filter mode.
        
        Args:
            bandwidth: filter bandwith, equals to control frequency(Hz)
        """
        cmode, imode = self._ctrl_mode_to_raw(MotorControlMode.POSITION_FILTER_MODE)
        return (cmode, imode, bandwidth)
    
    @send_to_bus(CMD_SET_MODE, '<BBeee')
    def position_track_mode(self, vel, acc, dec):
        """Enter position track mode.
        
        Args:
            vel: motion max vel(rad/s)
            acc: motion acceleration(rad/s^2)
            dec: motion deceleration(rad/s^2)
        """
        cmode, imode = self._ctrl_mode_to_raw(MotorControlMode.POSITION_TRACK_MODE)
        return (
            cmode, 
            imode, 
            fabs(vel * self._rot_factor), 
            fabs(acc * self._rot_factor), 
            fabs(dec * self._rot_factor)
        )
    
    @send_to_bus(CMD_SET_MODE, '<BB')
    def velocity_mode(self):
        """Enter velocity mode."""
        return self._ctrl_mode_to_raw(MotorControlMode.VELOCITY_MODE)
    
    @send_to_bus(CMD_SET_MODE, '<BBf')
    def velocity_ramp_mode(self, ramp):
        """Enter velocity ramp mode.
        
        Args:
            ramp: motion acceleration(rad/s^2)
        """
        cmode, imode = self._ctrl_mode_to_raw(MotorControlMode.VELOCITY_RAMP_MODE)
        return (cmode, imode, fabs(ramp * self._rot_factor))
    
    @send_to_bus(CMD_SET_MODE, '<BB')
    def torque_mode(self):
        """Enter torque mode"""
        return self._ctrl_mode_to_raw(MotorControlMode.TORQUE_MODE)
    
    @get_from_bus(CMD_GET_PID, '<fee')
    def get_pid(self, pos_kp, vel_kp, vel_ki):
        """Get motor pid params.
        
        Args:
            timeout: 0 indicates unlimited timeout(s)
        
        Returns:
            pos_kp, vel_kp, vel_ki
            None if timeout
        """
        return (pos_kp, vel_kp, vel_ki)
    
    @send_to_bus(CMD_SET_PID, '<fee')
    def set_pid(self, pos_kp, vel_kp, vel_ki):
        """Set motor pid params.
        
        Args:
            pos_kp: kp of position control
            vel_kp: kp of velocity control
            vel_ki: ki of velocity control
        """
        return (pos_kp, vel_kp, vel_ki)

    @get_from_bus(CMD_GET_LIMITS, '<ff')
    def get_vel_limit(self, vel_limit, current_limit):
        """Get motor volocity limit.
        
        Args:
            timeout: 0 indicates unlimited timeout(s)
        
        Returns:
            velocity limit(rad/s)
            None if timeout
        """
        return vel_limit / fabs(self._rot_factor)
    
    @get_from_bus(CMD_GET_LIMITS, '<ff')
    def get_current_limit(self, vel_limit, current_limit):
        """Get motor current limit.
        
        Args:
            timeout: 0 indicates unlimited timeout(s)
        
        Returns:
            current limit(A)
            None if timeout
        """
        return current_limit

    @send_to_bus(CMD_SET_LIMITS, '<ff')
    def set_vel_limit(self, vel_lim):
        """Set motor velocity limit.
        
        Args:
            vel_lim: velocity limit(rad/s)
        """
        return (fabs(vel_lim * self._rot_factor), 0)
    
    @send_to_bus(CMD_SET_LIMITS, '<ff')
    def set_current_limit(self, current_lim):
        """Set motor current limit.
        
        Args:
            current_lim: current limit(A)
        """
        return (0, current_lim)
    
    @send_to_bus(CMD_SET_POS, '<fee')
    def set_pos(self, pos, vel_ff, torque_ff):
        """Set motor target position.
        
        Args:
            pos: target position(rad)
            vel_ff: velocity feed forward(rad/s)
            torque_ff:torque feed forward(Nm)
        """
        return (pos * self._rot_factor, vel_ff * self._rot_factor, torque_ff / self._reduction)

    @send_to_bus(CMD_SET_VEL, '<ff')
    def set_vel(self, vel, torque_ff):
        """Set motor velocity.
        
        Args:
            vel: target velocity(rad)
            torque_ff: torque feed forward(Nm)
        """
        return (vel * self._rot_factor, torque_ff / self._reduction)
    
    @send_to_bus(CMD_SET_TORQUE, '<f')
    def set_torque(self, torque):
        """Set motor torque.
        
        Args:
            torque: target torque(Nm)
        """
        return (torque / self._reduction, )

    @send_to_bus(CMD_RESET)
    def reset(self):
        """Reset motor with factory configurations."""
        pass

    @send_to_bus(CMD_UNLOCK)
    def unlock(self):
        """Unlock motor brake if motor has one."""
        pass



import sys
import time
import argparse
import colorama
from termcolor import colored
sys.path.append("..")
from tools.list_devices import list_devices
from tools.monitor import monitor
from robodyno_components_python.interfaces import CanBus
from robodyno_components_python.components import Motor

colorama.init(autoreset=True)

parser = argparse.ArgumentParser(description='Robodyno command line tools.')
parser.add_argument('command', choices=['list', 'monitor', 'motor'], help='''
                        list all robodyno devices
                        / monitor messages on bus
                        / motor basic operations
                    ''')
parser.add_argument('--id', type=str, nargs='*', default=[], 
                    help='robodyno device id filter')
parser.add_argument('-c', '--channel', type=str, default='can0',
                    help='can bus channel')
parser.add_argument('-b', '--bitrate', type=str, default='CAN_1M',
                    help='can bus bitrate')

args = parser.parse_args()

def parse_id(str_list):
    """parse list of hex id from string format
    
    Args:
        str_list: string id list
    
    Returns:
        hex id list
    """
    if len(str_list) > 0 and str_list[0] == 'all':
        return [id for id in range(0x01, 0x40)]
    else:
        return list(map(lambda id_str: int(id_str, 0), str_list))

try:
    can = CanBus(bitrate = args.bitrate, channel = args.channel)
except:
    raise ValueError('Wrong can bus configurations.')

try:
    id_list = parse_id(args.id)
except:
    raise ValueError('Wrong id list.')

if args.command == 'list':
    list_devices(can, id_list)

if args.command == 'monitor':
    monitor(can, id_list)



#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""utils.py
Time    :   2022/10/17
Author  :   song 
Version :   1.0
Contact :   zhaosongy@126.com
License :   (C)Copyright 2022, robottime / robodyno

Can Bus Interface Utilities
"""

import time
from enum import Enum
import struct

class CanSpeed(Enum):
    CAN_1M    = 1000000
    CAN_1000K = 1000000
    CAN_500K  = 500000
    CAN_250K  = 250000

def get_from_bus(command_id, format):
    """Decorator for getting data from can bus.
    
    Args:
        command_id: command id
        format: python struct format to unpack msg bytes
    
    Returns:
        decorated function
    """
    def wrapper(func):
        def innerwrapper(self, timeout = 0):
            try:
                def callback(device_id, command_id, data, timestamp):
                    callback.data = data
                    callback.updated = True
                callback.updated = False

                self._iface().subscribe(
                    device_id = self.id, 
                    command_id = command_id, 
                    callback = callback
                )
                self._iface().send_can_msg(self.id, command_id, b'', True)

                start = time.time()
                while timeout == 0 or time.time() - start < timeout:
                    if callback.updated:
                        try:
                            return func(self, *struct.unpack(format, callback.data))
                        except:
                            raise ValueError()
                self._iface().unsubscribe(
                    device_id = self.id,
                    command_id = command_id,
                    callback = callback
                )
            except:
                raise RuntimeError('Failed to get data from can bus.')
        return innerwrapper
    return wrapper

def send_to_bus(command_id, format = '', remote = False):
    """Decorator for sending data from can bus.
    
    Args:
        command_id: command id
        format: python struct format to pack msg bytes
        remote: can frame rtr bit
    
    Returns:
        decorated function
    """
    def wrapper(func):
        def inner_wrapper(self, *args, **kwargs):
            try:
                data = b''
                if len(format) > 0:
                    data = struct.pack(format, *func(self, *args, **kwargs))
                self._iface().send_can_msg(self.id, command_id, data, remote)
            except:
                raise RuntimeError('Failed to send data from can bus.')
        return inner_wrapper
    return wrapper


#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""six_dof_collab_robot.py
Time    :   2022/10/08
Author  :   ryan 
Version :   1.0
Contact :   ryanzhang@163.com
License :   (C)Copyright 2022, robottime / robodyno

6DoF Collaborative Robot Driver

  Typical usage example:

  from robodyno.robot.six_dof_collaborative_robot import SixDoFCollabRobot
  robot = SixDoFCollabRobot(
      j1 = base_motor,
      j2 = shoulder_motor,
      j3 = upperarm_motor,
      j4 = elbow_motor,
      j5 = forearm_motor,
      j6 = hand_motor,
      l01 = .18,
      l12 = .135,
      l23 = .135,
      l34 = .075,
      l45 = .075,
      l56 = .1,
      end_effector = None
  )
"""

import time
from math import pi, sqrt, sin, cos, atan2
from cmath import acos, asin
import numpy as np
from ..utils.transformations import dh_matrix, translation_matrix, euler_matrix
from ..utils.interpolations import linear_interpolation

class SixDoFCollabRobot(object):
    """6 DoF collaborarive robot driver
    
    Attributes:
        joints: list of 6 joint motors
        l01: link from world to joint 1
        l12: link from joint 1 to joint 2
        l23: link from joint 2 to joint 3
        l34: link from joint 3 to joint 4
        l45: link from joint 4 to joint 5
        l56: link from joint 5 to joint 6
        end_effector: end effector object
    """
    
    def __init__(self, j1, j2, j3, j4, j5, j6, l01, l12, l23, l34, l45, l56, end_effector = None):
        """init robot with joints and links"""
        self.joints = [j1, j2, j3, j4, j5, j6]
        self.l01 = l01
        self.l12 = l12
        self.l23 = l23
        self.l34 = l34
        self.l45 = l45
        self.l56 = l56
        self.end_effector = end_effector

        for m in self.joints:
            m.position_filter_mode(8)
        self._axes_poses = [0 for i in range(6)]
        self._axes_zeros = [0 for i in range(6)]

    def get_joints_poses(self):
        """Read joints positions to a list.
        
        Returns:
            a list of 6 joints positions
        """
        poses = []
        for i in range(6):
            pos = None
            count = 0
            while not pos:
                if count > 5:
                    raise RuntimeError('Filed to get position of Joint {}.'.format(i+1))
                count += 1
                pos = self.joints[i].get_pos(0.2)
            poses.append(pos - self._axes_zeros[i])
        return poses

    def enable(self):
        for i in range(6):
            self.joints[i].enable()
    
    def disable(self):
        for i in range(6):
            self.joints[i].disable()
    
    def init(self, axes_poses = [0 for i in range(6)]):
        """Calibrate robot motors with given axes poses.
        
        Args:
            axes_poses: list of 6 axes poses(rad)
        """
        self._axes_poses = axes_poses.copy()
        self._axes_zeros = [0 for i in range(6)]
        cur_poses = self.get_joints_poses()
        for i in range(6):
            self._axes_zeros[i] = cur_poses[i] - self._axes_poses[i]

    def set_joint(self, id, pos):
        """Set joint angle with joint id and position
        
        Args:
            id: joint id (0-5)
            pos: joint target position(rad)
        """
        self.joints[id].set_pos(pos + self._axes_zeros[id])
        self._axes_poses[id] = pos
    
    def joint_space_interpolated_motion(self, target, speeds = [None for i in range(6)], duration = 0):
        """Robot interpolated motion in joint space.
        
        Args:
            target: iterable of 6 joints target angle(rad)
            speeds: iterable of 6 joints motion speed(rad/s)
            duration: default motion duration(s)
        """
        interpolations = []
        joint_poses = self.get_joints_poses()
        for i in range(6):
            interpolations.append(linear_interpolation(joint_poses[i], target[i], speed = speeds[i], duration=duration))
        update_flag = True
        while update_flag:
            update_flag = False
            for i in range(6):
                pos = next(interpolations[i], None)
                if pos is not None:
                    self.set_joint(i, pos)
                    update_flag = True
            time.sleep(0.05)

    def home(self, duration = 5):
        """Move back to zero position.
        
        Args:
            duration: motion duration(s)
        """
        self.joint_space_interpolated_motion((0,0,0,0,0,0), duration = duration)

    def cartesian_space_interpolated_motion(self, x, y, z, roll, pitch, yaw, sol_id = 0, 
        x_speed = None, y_speed = None, z_speed = None, roll_speed = None, 
        pitch_speed = None, yaw_speed = None, duration = 0):
        """Robot Interpolated motion in Cartesian space.
        
        Args:
            x: target robot end x
            y: target robot end y
            z: target robot end z
            roll: target robot end roll
            pitch: target robot end pitch
            yaw: target robot end yaw
            sol_id: inverse kinematics solve id (0-7)
            x_speed: speed alone X dimetion(m/s)
            y_speed: speed alone Y dimetion(m/s)
            z_speed: speed alone Z dimetion(m/s)
            roll_speed: rotation speed on X axis(rad/s)
            pitch_speed: rotation speed on Y axis(rad/s)
            yaw_speed: rotation speed on Z axis(rad/s)
            duration: default motion duration(s)
        """
        current_pose = self.forward_kinematics(self.get_joints_poses())
        current_x = current_pose['x']
        current_y = current_pose['y']
        current_z = current_pose['z']
        current_roll = current_pose['roll']
        current_pitch = current_pose['pitch']
        current_yaw = current_pose['yaw']

        x_interpolation = linear_interpolation(current_x, x, x_speed, duration)
        y_interpolation = linear_interpolation(current_y, y, y_speed, duration)
        z_interpolation = linear_interpolation(current_z, z, z_speed, duration)
        roll_interpolation = linear_interpolation(current_roll, roll, roll_speed, duration)
        pitch_interpolation = linear_interpolation(current_pitch, pitch, pitch_speed, duration)
        yaw_interpolation = linear_interpolation(current_yaw, yaw, yaw_speed, duration)
        while True:
            temp_x = next(x_interpolation, x)
            temp_y = next(y_interpolation, y)
            temp_z = next(z_interpolation, z)
            temp_roll = next(roll_interpolation, roll)
            temp_pitch = next(pitch_interpolation, pitch)
            temp_yaw = next(yaw_interpolation, yaw)
            axes = self.inverse_kinematics(temp_x, temp_y, temp_z, temp_roll, temp_pitch, temp_yaw, sol_id)
            for i in range(6):
                self.set_joint(i, axes[i])
            if temp_x == x and temp_y == y and temp_z == z and temp_roll == roll and temp_pitch == pitch and temp_yaw == yaw:
                break
            time.sleep(0.05)

    def forward_kinematics(self, axes):
        """Forward kinematics algorism
        
        Args:
            axes: list of 6 joint angles
        
        Returns:
            dictionary of robot end transform
            
            example:
            {
                'x': 0,
                'y': 0,
                'z': 0,
                'roll': 0,
                'pitch': 0,
                'yaw': 0,
            }
        """
        T01 = dh_matrix(pi/2 , 0 , self.l01, axes[0])
        T12 = dh_matrix(0    , -self.l12, 0 , axes[1])
        T23 = dh_matrix(0    , -self.l23, 0 , axes[2])
        T34 = dh_matrix(pi/2 , 0 , self.l34, axes[3])
        T45 = dh_matrix(-pi/2, 0 , self.l45, axes[4])
        T56 = dh_matrix(0    , 0 , self.l56, axes[5])
        T06 = T01 @ T12 @ T23 @ T34 @ T45 @ T56
        return T06

    def inverse_kinematics(self, x, y, z, roll, pitch, yaw, sol_id):
        """Inverse kinematics algorism
        
        Args:
            x: robot end x
            y: robot end y
            z: robot end z
            roll: robot end roll
            pitch: robot end pitch
            yaw: robot end yaw
        
        Returns:
            list of joint angles
        """
        sol = np.zeros((6, 8))
        T06 = translation_matrix([x,y,z]) @ euler_matrix(roll, pitch, yaw)
        P05 = T06 @ np.array([[0, 0, -self.l56, 1]]).T
        
        # solve theta1
        phi1 = atan2(P05[1, 0], P05[0, 0])
        phi2 = acos(self.l34 / sqrt(P05[1, 0]*P05[1, 0] + P05[0, 0] * P05[0, 0])).real
        sol[0, 0:4] = pi / 2 + phi1 + phi2
        sol[0, 4:8] = pi / 2 + phi1 - phi2
        
        for i in range(8):
            # solve theta5
            sol[4, i] = acos((T06[0, 3] * sin(sol[0, i]) - T06[1, 3] * cos(sol[0, i]) - self.l34) / self.l56).real
            if i % 2 == 1:
                sol[4, i] = -sol[4, i]
            
            # solve theta6
            T01 = dh_matrix(pi / 2, 0, self.l01, sol[0, i])
            T10 = np.linalg.inv(T01)
            T61 = np.linalg.inv(T10 @ T06)
            sol[5, i] = atan2(-T61[1, 2] / sin(sol[4, i]), T61[0, 2] / sin(sol[4, i]))

            # solve theta3
            T45 = dh_matrix(-pi / 2, 0, self.l45, sol[4, i])
            T56 = dh_matrix(0, 0, self.l56, sol[5, i])
            T14 = T10 @ T06 @ np.linalg.inv(T56) @ np.linalg.inv(T45)
            P13 = T14 @ np.array([[0, -self.l34, 0, 1]]).T - np.array([[0, 0, 0, 1]]).T
            P13_norm = np.linalg.norm(P13)
            sol[2, i] = acos((P13_norm * P13_norm - self.l12 * self.l12 - self.l23 * self.l23) / (2 * self.l12 * self.l23)).real
            if i == 2 or i == 3 or i == 6 or i == 7:
                sol[2, i] = -sol[2, i]

            # solve theta 2
            sol[1, i] = atan2(-P13[1, 0], -P13[0, 0]) - asin((self.l23 * sin(sol[2, i])) / P13_norm).real

            # solve theta 4
            T12 = dh_matrix(0, -self.l12, 0, sol[1, i])
            T23 = dh_matrix(0, -self.l23, 0, sol[2, i])
            T34 = np.linalg.inv(T23) @ np.linalg.inv(T12) @ T14
            sol[3, i] = atan2(T34[1, 0], T34[0, 0])

        axes = []
        for i in range(6):
            axes.append(sol[i, sol_id])
        axes[1] += pi / 2
        axes[2] = -axes[2]
        axes[3] += pi / 2
        return axes



#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""device_node.py
Time    :   2022/10/17
Author  :   song 
Version :   1.0
Contact :   zhaosongy@126.com
License :   (C)Copyright 2022, robottime / robodyno
"""

from ...components.device import *
from .utils import get_from_bus, send_to_bus

class CanBusDeviceNode(DeviceNode):
    """Can Bus Robodyno Common Device Node"""

    CMD_GET_VERSION = 0x01

    def __init__(self, device, iface):
        """Init node from robodyno common device and can bus interface
        
        Args:
            device: robodyno common device object
            iface: can bus interface
        """
        super().__init__(device, iface)
        self.id = device.id
    
    @get_from_bus(CMD_GET_VERSION, '<HHI')
    def get_version(self, main_ver, sub_ver, type):
        """Get device firmware / simulation version.
        
        Args:
            timeout: 0 indicates unlimited timeout(s)
        
        Returns:
            dictionary of device version
            None if timeout
        """
        return {
            'main_version': main_ver,
            'sub_version': sub_ver,
            'type': DeviceType(type)
        }
    


#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""device.py
Time    :   2022/10/17
Author  :   song 
Version :   1.0
Contact :   zhaosongy@126.com
License :   (C)Copyright 2022, robottime / robodyno

Robodyno common device object

  Typical usage example:

  from robodyno.interfaces import CanBus
  from robodyno.components import Device

  can = CanBus()
  device = Device(iface = can, id = 0x10)
  device.get_version()
"""

from enum import Enum
from abc import ABC
import weakref

class DeviceType(Enum):
    ROBODYNO_PRO_44 = 0x00
    ROBODYNO_PRO_12 = 0x01
    ROBODYNO_PLUS_50  = 0x10
    ROBODYNO_PLUS_100 = 0x11
    ROBODYNO_EXB_FCTY = 0x80

class DeviceNode(ABC):
    """Robodyno common device node abstract class"""

    def __init__(self, device, iface):
        """Init node with device object and interface.
        
        Args:
            device: robodyno device object
            iface: robodyno interface
        """
        self._device = weakref.ref(device)
        self._iface = weakref.ref(iface)
    
    def _delete(self):
        """Collect node from memory."""
        pass

    def get_version(self, timeout):
        """Get device firmware / simulation version.
        
        Args:
            timeout: 0 indicates unlimited timeout(s)
        
        Returns:
            dictionary of motor version
            None if timeout
        """
        return None

class Device(object):
    """Robodyno common device
    
    Attributes:
        id: correspond with device id in real world or simulation.
    """
    
    def __init__(self, iface, id = 0x10):
        """Init device with interface and id
        
        Args:
            iface: robodyno interface
            id: device id
        """
        try:
            self.id = id
            if id < 0x01 or id >= 0x40:
                raise ValueError
        except:
            raise ValueError('Device id is not available.')
        
        try:
            self._node = iface.get_device_node(self)
        except:
            raise RuntimeError('Failed to init motor from interface.')
    
    def get_version(self, timeout = 0):
        """Blocking function to get device firmware / simulation version.
        
        Args:
            timeout: 0 indicates unlimited timeout(s)
        
        Returns:
            dictionary of device version
            None if timeout
        """
        try:
            return self._node.get_version(timeout)
        except:
            raise RuntimeError('Failed to invoke device api from interface.')



#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""list.py
Time    :   2022/10/17
Author  :   song 
Version :   1.0
Contact :   zhaosongy@126.com
License :   (C)Copyright 2022, robottime / robodyno

List robodyno devices on can bus.
"""

from termcolor import colored
from ..components import Device

def get_device_version(bus, id):
    """Read robodyno device version from bus with id
    
    Args:
        bus: robodyno can bus
        id: robodyno device id
    
    Returns:
        robodyno device version dict
    """
    device = Device(bus, id)
    return device.get_version(0.05)

def display_ver(id, ver):
    """Display robodyno device version
    
    Args:
        id: robodyno device id
        ver: robodyno device version dict
    """
    print(colored('[0x{:02X}]'.format(id), 'cyan') + ' ' +
          colored(ver['type'].name, 'green') + ', ' +
          'ver: {}.{}'.format(ver['main_version'], ver['sub_version']))

def list_devices(can, id_list):
    """Print all device and version with id selected on can bus.
    
    Args:
        can: can bus interface
        id_list: selected id list
    """
    if len(id_list) > 0:
        for id in id_list:
            ver = get_device_version(can, id)
            if ver:
                display_ver(id, ver)
    else:
        for id in range(1, 0x40):
            ver = get_device_version(can, id)
            if ver:
                display_ver(id, ver)
    can.disconnect()


import sys

if sys.platform == 'win32':
    from .can_bus.candle_interface import CandleInterface as CanBus
elif sys.platform.startswith('linux'):
    from .can_bus.socketcan_interface import SocketCanInterface as CanBus




#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""motor.py
Time    :   2022/09/30
Author  :   song 
Version :   1.0
Contact :   zhaosongy@126.com
License :   (C)Copyright 2022, robottime / robodyno

Robodyno Motor Python Driver

  Typical usage example:

  from robodyno.interfaces import CanBus, Webots
  from robodyno.components import Motor

  can = CanBus()
  webots = Webots()
  
  motor_can = Motor(iface = can, id = 0x10, type = 'ROBODYNO_PRO_44')
  motor_webots = Motor(iface = webots, id = 0x11, type = 'ROBODYNO_PRO_12')

  can.disconnect()
"""

from enum import Enum
from abc import ABC
import weakref
import time

class MotorApiRuntimeError(RuntimeError):
    def __init__(self):
        super().__init__('Failed to invoke motor api from interface.')

class MotorType(Enum):
    ROBODYNO_PRO_44 = 0x00
    ROBODYNO_PRO_12 = 0x01
    ROBODYNO_PLUS_50  = 0x10
    ROBODYNO_PLUS_100 = 0x11

class MotorState(Enum):
    UNKNOWN = -1
    DISABLED = 1
    CALIBRATE = 3
    MOTOR_CALIBRATING = 4
    OFFSET_CALIBRATING = 7
    ENABLED = 8

class MotorError(Enum):
    NONE = 0
    # INVALID_STATE = 0 TODO: 修复固件发送错误值可能为0的bug
    UNDER_VOLTAGE = 1
    OVER_VOLTAGE = 2
    CURRENT_MEASUREMENT_TIMEOUT = 3
    BRAKE_RESISTOR_DISARMED = 4
    MOTOR_DISARMED = 5
    MOTOR_FAILED = 6
    SENSORLESS_ESTIMATOR_FAILED = 7
    ENCODER_FAILED = 8
    CONTROLLER_FAILED = 9
    POS_CTRL_DURING_SENSORLESS = 10
    WATCHDOG_TIMER_EXPIRED = 11
    MIN_ENDSTOP_PRESSED = 12
    MAX_ENDSTOP_PRESSED = 13
    ESTOP_REQUESTED = 14
    HOMING_WITHOUT_ENDSTOP = 17
    OVER_TEMP = 18

class MotorMotorError(Enum):
    NONE = 0
    # PHASE_RESISTANCE_OUT_OF_RANGE = 0
    PHASE_INDUCTANCE_OUT_OF_RANGE = 1
    ADC_FAILED = 2
    DRV_FAULT = 3
    CONTROL_DEADLINE_MISSED = 4
    NOT_IMPLEMENTED_MOTOR_TYPE = 5
    BRAKE_CURRENT_OUT_OF_RANGE = 6
    MODULATION_MAGNITUDE = 7
    BRAKE_DEADTIME_VIOLATION = 8
    UNEXPECTED_TIMER_CALLBACK = 9
    CURRENT_SENSE_SATURATION = 10
    CURRENT_LIMIT_VIOLATION = 12
    BRAKE_DUTY_CYCLE_NAN = 13
    DC_BUS_OVER_REGEN_CURRENT = 14
    DC_BUS_OVER_CURRENT = 15

class MotorEncoderError(Enum):
    NONE = 0
    # UNSTABLE_GAIN = 0
    CPR_POLEPAIRS_MISMATCH = 1
    NO_RESPONSE = 2
    UNSUPPORTED_ENCODER_MODE = 3
    ILLEGAL_HALL_STATE = 4
    INDEX_NOT_FOUND_YET = 5
    ABS_SPI_TIMEOUT = 6
    ABS_SPI_COM_FAIL = 7
    ABS_SPI_NOT_READY = 8
    
class MotorControllerError(Enum):
    NONE = 0
    # OVERSPEED = 0
    INVALID_INPUT_MODE = 1
    UNSTABLE_GAIN = 2
    INVALID_MIRROR_AXIS = 3
    INVALID_LOAD_ENCODER = 4
    INVALID_ESTIMATE = 5

class MotorControlMode(Enum):
    UNKNOWN = -1
    POSITION_MODE = 0
    POSITION_FILTER_MODE = 1
    POSITION_TRACK_MODE = 2
    VELOCITY_MODE = 3
    VELOCITY_RAMP_MODE = 4
    TORQUE_MODE = 5

class MotorNode(ABC):
    """Robodyno motor node abstract class."""
    
    def __init__(self, motor, iface):
        """Init node with motor object and robodyno interface.
        
        Args:
            motor: motor object
            iface: robodyno interface
        """
        self._motor = weakref.ref(motor)
        self._iface = weakref.ref(iface)
    
    def _delete(self):
        """Collect node from memory."""
        pass        
    
    def get_version(self, timeout):
        """Get motor firmware / simulation version.
        
        Args:
            timeout: 0 indicates unlimited timeout(s)
        
        Returns:
            dictionary of motor version
            None if timeout
        """
        return None
    
    def estop(self):
        """Emergency stop motor."""
        pass
    
    def reboot(self):
        """Reboot motor."""
        pass

    def clear_errors(self):
        """Try to clear motor errors."""
        pass
        
    def save_configuration(self):
        """Save configurations to hardware."""
        pass

    def config_can_bus(self, new_id, heartbeat, bitrate):
        """Configure motor can bus settings.
        
        Args:
            new_id: motor device id(0x01-0x3f)
            heartbeat: heartbeat period(ms)
            bitrate: motor can bus bitrate(save and reboot to take effect)
        """
        pass

    def set_state(self, state):
        """Cange motor state.
        
        Args:
            state: motor state(MotorState)
        """
        pass

    def get_voltage(self, timeout):
        """Get motor vbus voltage.
        
        Args:
            timeout: 0 indicates unlimited timeout(s)
        
        Returns:
            voltage(V)
            None if timeout
        """
        return None
    
    def get_temperature(self, timeout):
        """Get motor temperature.
        
        Args:
            timeout: 0 indicates unlimited timeout(s)
        
        Returns:
            temperature(°C)
            None if timeout
        """
        return None
    
    def get_feedback(self, timeout):
        """Get motor position, velocity and torque feedback.
        
        Args:
            timeout: 0 indicates unlimited timeout(s)
        
        Returns:
            position(rad), velocity(rad/s), torque(Nm)
            None if timeout
        """
        return None
    
    def get_mode(self, timeout):
        """Get motor control mode and params.
        
        Args:
            timeout: 0 indicates unlimited timeout(s)
        
        Returns:
            control mode(MotorControlMode), control params dict
            None if timeout
        """
        return None
        
    def position_mode(self):
        """Enter position pid mode."""
        pass

    def position_filter_mode(self, bandwidth):
        """Enter position filter mode.
        
        Args:
            bandwidth: filter bandwith, equals to control frequency(Hz)
        """
        pass

    def position_track_mode(self, vel, acc, dec):
        """Enter position track mode.
        
        Args:
            vel: motion max vel(rad/s)
            acc: motion acceleration(rad/s^2)
            dec: motion deceleration(rad/s^2)
        """
        pass

    def velocity_mode(self):
        """Enter velocity mode."""
        pass

    def velocity_ramp_mode(self, ramp):
        """Enter velocity ramp mode.
        
        Args:
            ramp: motion acceleration(rad/s^2)
        """
        pass

    def torque_mode(self):
        """Enter torque mode"""
        pass

    def get_pid(self, timeout):
        """Get motor pid params.
        
        Args:
            timeout: 0 indicates unlimited timeout(s)
        
        Returns:
            pos_kp, vel_kp, vel_ki
            None if timeout
        """
        return None
    
    def set_pid(self, pos_kp, vel_kp, vel_ki):
        """Set motor pid params.
        
        Args:
            pos_kp: kp of position control
            vel_kp: kp of velocity control
            vel_ki: ki of velocity control
        """
        pass

    def get_vel_limit(self, timeout):
        """Get motor volocity limit.
        
        Args:
            timeout: 0 indicates unlimited timeout(s)
        
        Returns:
            velocity limit(rad/s)
            None if timeout
        """
        return None
        
    def get_current_limit(self, timeout):
        """Get motor current limit.
        
        Args:
            timeout: 0 indicates unlimited timeout(s)
        
        Returns:
            current limit(A)
            None if timeout
        """
        return None

    def set_vel_limit(self, vel_lim):
        """Set motor velocity limit.
        
        Args:
            vel_lim: velocity limit(rad/s)
        """
        pass

    def set_current_limit(self, current_lim):
        """Set motor current limit.
        
        Args:
            current_lim: current limit(A)
        """
        pass

    def set_pos(self, pos, vel_ff, torque_ff):
        """Set motor target position.
        
        Args:
            pos: target position(rad)
            vel_ff: velocity feed forward(rad/s)
            torque_ff:torque feed forward(Nm)
        """
        pass

    def set_vel(self, vel, torque_ff):
        """Set motor velocity.
        
        Args:
            vel: target velocity(rad)
            torque_ff: torque feed forward(Nm)
        """
        pass

    def set_torque(self, torque):
        """Set motor torque.
        
        Args:
            torque: target torque(Nm)
        """
        pass

    def reset(self):
        """Reset motor with factory configurations."""
        pass

    def unlock(self):
        """Unlock motor brake if motor has one."""
        pass

class Motor(object):
    """Robodyno Motor
    
    Attributes:
        id: correspond with motor id in real world or simulation.
        type: motor type enum
        reduction: motor reduction
        available_volocity: theoretical max volocity(rad/s)
        available_torque: theoretical max torque(Nm)
        available_current: theoretical max phase current(A)
        torque_constant: theoretical torque provided by unit current(Nm/A)
        with_brake: is motor has a brake(bool)
        state: motor state read from heartbeat pack
        error: motor error read from heartbeat pack
        mode: motor mode read from heartbeat pack
        sync_time: last timestamp when receive heartbeat pack
    """

    def __init__(self, iface, id = 0x10, type = None):
        """Init motor from interface, id and type.
        
        Args:
            iface: robodyno interface object
            id: range from 0x01 to 0x40
            type: Motor type string
        """
        try:
            self.id = id
            if id < 0x01 or id >= 0x40:
                raise ValueError
        except:
            raise ValueError('Motor id is not available.')

        try:
            self.type = MotorType[type]
        except:
            try:
                m = Motor(iface, id, MotorType(0).name)
                ver = m.get_version(timeout=0.05)
                self.type = ver['type']
            except:
                raise ValueError('Motor type is not set and can not recognized automatically.')

        self.__dict__.update(self.__get_spec())

        try:
            self._node = iface.get_motor_node(self)
        except:
            raise RuntimeError('Failed to init motor from interface.')

        
        # TODO: 添加同步心跳包的功能
        self.state = MotorState.UNKNOWN
        self.error = {
            'error': MotorError.NONE,
            'motor_err': MotorMotorError.NONE,
            'encoder_err': MotorEncoderError.NONE,
            'controller_err': MotorControllerError.NONE,
        }
        self.mode = MotorControlMode.UNKNOWN
        self.sync_time = 0
    
    def __del__(self):
        try:
            self._node._delete()
        except:
            pass

    def __get_spec(self):
        """get motor factory specifications"""
        robodyno_motor_specs = {
            MotorType.ROBODYNO_PRO_44: {
                'reduction': -44,
                'available_velocity': 27,
                'available_torque': 13,
                'available_current': 15,
                'torque_constant': 0.7742,
                'with_brake': False,
            },
            MotorType.ROBODYNO_PRO_12: {
                'reduction': -12.45,
                'available_velocity': 95,
                'available_torque': 4,
                'available_current': 15,
                'torque_constant': 0.2191,
                'with_brake': False,
            },
            MotorType.ROBODYNO_PLUS_50: {
                'reduction': 50,
                'available_velocity': 24,
                'available_torque': 48,
                'available_current': 25,
                'torque_constant': 1.7229,
                'with_brake': True,
            },
            MotorType.ROBODYNO_PLUS_100: {
                'reduction': 100,
                'available_velocity': 12,
                'available_torque': 96,
                'available_current': 25,
                'torque_constant': 3.4458,
                'with_brake': True,
            },
        }
        return robodyno_motor_specs.get(self.type, None)

    def get_version(self, timeout = 0):
        """Blocking function to get motor firmware / simulation version.
        
        Args:
            timeout: 0 indicates unlimited timeout(s)
        
        Returns:
            dictionary of motor version
            None if timeout
        """
        try:
            return self._node.get_version(timeout)
        except:
            raise MotorApiRuntimeError
    
    def estop(self):
        """Stop motor by set an error bit, reboot motor to recover."""
        try:
            self._node.estop()
        except:
            raise MotorApiRuntimeError
    
    def reboot(self):
        """Reboot motor."""
        try:
            self._node.reboot()
        except:
            raise MotorApiRuntimeError
    
    def clear_errors(self):
        """Try to clear motor errors."""
        try:
            self._node.clear_errors()
        except:
            raise MotorApiRuntimeError
    
    def save_configuration(self):
        """Save motor configurations."""
        try:
            self._node.save_configuration()
        except:
            raise MotorApiRuntimeError
    
    def config_can_bus(self, new_id, heartbeat = 1, bitrate = 'CAN_1M'):
        """Configure motor can bus settings.
        
        Args:
            new_id: motor device id(0x01-0x3f)
            heartbeat: heartbeat period(s)
            bitrate: choose from 'CAN_1000K', 'CAN_500K', 'CAN_250K',
                     (save and reboot to take effect)
        """
        if new_id < 0x01 or new_id > 0x3f:
            raise ValueError('Unavailable can id. Choose from 0x01-0x3f')
        try:
            self._node.config_can_bus(new_id, heartbeat, bitrate)
        except:
            raise MotorApiRuntimeError

    def enable(self):
        """Enable motor. Set motor state to CLOSED_LOOP."""
        try:
            self._node.set_state(MotorState.ENABLED)
            time.sleep(0.01)
        except:
            raise MotorApiRuntimeError
    
    def calibrate(self):
        """Calibrate motor. Set motor state to CALIBRATE."""
        try:
            self._node.set_state(MotorState.CALIBRATE)
        except:
            raise MotorApiRuntimeError
    
    def disable(self):
        """Disable motor. Set motor state to IDLE."""
        try:
            self._node.set_state(MotorState.DISABLED)
        except:
            raise MotorApiRuntimeError
        
    def get_voltage(self, timeout = 0):
        """Get motor vbus voltage.
        
        Args:
            timeout: 0 indicates unlimited timeout(s)
        
        Returns:
            voltage(V)
            None if timeout
        """
        try:
            return self._node.get_voltage(timeout)
        except:
            raise MotorApiRuntimeError
        
    def get_temperature(self, timeout = 0):
        """Get motor temperature.
        
        Args:
            timeout: 0 indicates unlimited timeout(s)
        
        Returns:
            temperature(°C)
            None if timeout
        """
        try:
            return self._node.get_temperature(timeout)
        except:
            raise MotorApiRuntimeError
    
    def get_feedback(self, timeout = 0):
        """Get motor position, velocity and torque feedback.
        
        Args:
            timeout: 0 indicates unlimited timeout(s)
        
        Returns:
            position(rad), velocity(rad/s), torque(Nm)
            None if timeout
        """
        try:
            return self._node.get_feedback(timeout)
        except:
            raise MotorApiRuntimeError
    
    def get_pos(self, timeout = 0):
        """Get motor position.
        
        Args:
            timeout: 0 indicates unlimited timeout(s)
        
        Returns:
            position(rad)
            None if timeout
        """
        try:
            return self._node.get_feedback(timeout)[0]
        except:
            raise MotorApiRuntimeError
    
    def get_vel(self, timeout = 0):
        """Get motor velocity.
        
        Args:
            timeout: 0 indicates unlimited timeout(s)
        
        Returns:
            velocity(rad/s)
            None if timeout
        """
        try:
            return self._node.get_feedback(timeout)[1]
        except:
            raise MotorApiRuntimeError
    
    def get_torque(self, timeout = 0):
        """Get motor torque.
        
        Args:
            timeout: 0 indicates unlimited timeout(s)
        
        Returns:
            torque(Nm)
            None if timeout
        """
        try:
            return self._node.get_feedback(timeout)[2]
        except:
            raise MotorApiRuntimeError
    
    def get_mode(self, timeout = 0):
        """Get motor control mode and params.
        
        Args:
            timeout: 0 indicates unlimited timeout(s)
        
        Returns:
            control mode(MotorControlMode), control params dict
            None if timeout
        """
        try:
            return self._node.get_mode(timeout)
        except:
            raise MotorApiRuntimeError
        
    def position_mode(self):
        """Enter position pid mode."""
        try:
            self._node.position_mode()
        except:
            raise MotorApiRuntimeError

    def position_filter_mode(self, bandwidth):
        """Enter position filter mode.
        
        Args:
            bandwidth: filter bandwith, equals to control frequency(Hz)
        """
        try:
            self._node.position_filter_mode(bandwidth)
        except:
            raise MotorApiRuntimeError

    def position_track_mode(self, vel, acc, dec):
        """Enter position track mode.
        
        Args:
            vel: motion max vel(rad/s)
            acc: motion acceleration(rad/s^2)
            dec: motion deceleration(rad/s^2)
        """
        try:
            self._node.position_track_mode(vel, acc, dec)
        except:
            raise MotorApiRuntimeError

    def velocity_mode(self):
        """Enter velocity mode."""
        try:
            self._node.velocity_mode()
        except:
            raise MotorApiRuntimeError

    def velocity_ramp_mode(self, ramp):
        """Enter velocity ramp mode.
        
        Args:
            ramp: motion acceleration(rad/s^2)
        """
        try:
            self._node.velocity_ramp_mode(ramp)
        except:
            raise MotorApiRuntimeError

    def torque_mode(self):
        """Enter torque mode"""
        try:
            self._node.torque_mode()
        except:
            raise MotorApiRuntimeError

    def get_pid(self, timeout = 0):
        """Get motor pid params.
        
        Args:
            timeout: 0 indicates unlimited timeout(s)
        
        Returns:
            pos_kp, vel_kp, vel_ki
            None if timeout
        """
        try:
            return self._node.get_pid(timeout)
        except:
            raise MotorApiRuntimeError
    
    def set_pid(self, pos_kp, vel_kp, vel_ki):
        """Set motor pid params.
        
        Args:
            pos_kp: kp of position control
            vel_kp: kp of velocity control
            vel_ki: ki of velocity control
        """
        try:
            self._node.set_pid(pos_kp, vel_kp, vel_ki)
        except:
            raise MotorApiRuntimeError

    def get_vel_limit(self, timeout = 0):
        """Get motor volocity limit.
        
        Args:
            timeout: 0 indicates unlimited timeout(s)
        
        Returns:
            velocity limit(rad/s)
            None if timeout
        """
        try:
            return self._node.get_vel_limit(timeout)
        except:
            raise MotorApiRuntimeError
        
    def get_current_limit(self, timeout = 0):
        """Get motor current limit.
        
        Args:
            timeout: 0 indicates unlimited timeout(s)
        
        Returns:
            current limit(A)
            None if timeout
        """
        try:
            return self._node.get_current_limit(timeout)
        except:
            raise MotorApiRuntimeError

    def set_vel_limit(self, vel_lim):
        """Set motor velocity limit.
        
        Args:
            vel_lim: velocity limit(rad/s)
        """
        try:
            self._node.set_vel_limit(vel_lim)
        except:
            raise MotorApiRuntimeError

    def set_current_limit(self, current_lim):
        """Set motor current limit.
        
        Args:
            current_lim: current limit(A)
        """
        try:
            self._node.set_current_limit(current_lim)
        except:
            raise MotorApiRuntimeError

    def set_pos(self, pos, vel_ff = 0, torque_ff = 0):
        """Set motor target position.
        
        Args:
            pos: target position(rad)
            vel_ff: velocity feed forward(rad/s)
            torque_ff:torque feed forward(Nm)
        """
        try:
            self._node.set_pos(pos, vel_ff, torque_ff)
        except:
            raise MotorApiRuntimeError

    def set_vel(self, vel, torque_ff = 0):
        """Set motor velocity.
        
        Args:
            vel: target velocity(rad)
            torque_ff: torque feed forward(Nm)
        """
        try:
            self._node.set_vel(vel, torque_ff)
        except:
            raise MotorApiRuntimeError

    def set_torque(self, torque):
        """Set motor torque.
        
        Args:
            torque: target torque(Nm)
        """
        try:
            self._node.set_torque(torque)
        except:
            raise MotorApiRuntimeError

    def reset(self):
        """Reset motor with factory configurations."""
        try:
            self._node.reset()
        except:
            raise MotorApiRuntimeError

    def unlock(self):
        """Unlock motor brake if motor has one."""
        if not self.with_brake:
            raise RuntimeError('Motor do not have a brake.')
        try:
            self._node.reset()
        except:
            raise MotorApiRuntimeError



#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""candle_interface.py
Time    :   2022/09/30
Author  :   song 
Version :   1.0
Contact :   zhaosongy@126.com
License :   (C)Copyright 2022, robottime / robodyno

Candle Can Bus Interface(Windows)

  Typical usage example:

  can = CandleInterface(bitrate = 'CAN_1000K', auto_connect = True)
  can.disconnect()

  or 

  can = CandleInterface(bitrate = 'CAN_250K', auto_connect = False)
  can.connect()
  can.disconnect()
"""

import candle_driver as candle
from .can_bus_interface import CanBusInterface
from .utils import CanSpeed

class CandleInterface(CanBusInterface):
    """Candle driver interface, only available on Windows."""

    def __init__(self, bitrate = 'CAN_1000K', auto_connect = True, *args, **kwargs):
        """Init candle driver interface
        
        Args:
            bitrate: choose from 'CAN_1000K', 'CAN_500K', 'CAN_250K'
            auto_connect: if set to False, you need to call connect() manually
        """
        try:
            self._device = candle.list_devices()[0]
            self._channel = self._device.channel(0)
        except:
            raise IOError('Failed to open can bus.')
        
        try:
            self._bitrate = CanSpeed[bitrate].value
        except:
            raise ValueError('Bitrate is not available.')
        
        super().__init__(auto_connect)
    
    def _enable(self):
        """Connect to candle can bus."""
        if not self._device.open():
            raise IOError
        self._channel.set_bitrate(self._bitrate)
        if not self._channel.start():
            raise IOError
    
    def _disable(self):
        """Disconnect to candle can bus."""
        self._channel.stop()
        self._device.close()
    
    def _send(self, can_frame_id, data, remote):
        """Send can frame from candle interface."""
        if remote:
            can_frame_id |= candle.CANDLE_ID_RTR
        try:
            self._channel.write(can_frame_id, data)
        except:
            raise IOError('Failed to send to candle can bus.')

    def _read(self, timeout):
        """Read can frame from candle interface."""
        try:
            frame_type, frame_id, data, _, timestamp = self._channel.read(int(timeout * 1000))
            if frame_type == candle.CANDLE_FRAMETYPE_RECEIVE:
                return frame_id, data, timestamp
            else:
                return None
        except Exception as e:
            raise IOError('Failed to read from candle can bus.')



#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""socketcan_interface.py
Time    :   2022/10/12
Author  :   song 
Version :   1.0
Contact :   zhaosongy@126.com
License :   (C)Copyright 2022, robottime / robodyno

Socket Can Bus Interface(Windows)

  Typical usage example:

  can = SocketCanInterface(channel = 'can0', auto_connect = True)
  can.disconnect()

  or 

  can = SocketCanInterface(auto_connect = False)
  can.connect()
  can.disconnect()
"""

import can
from .can_bus_interface import CanBusInterface

class SocketCanInterface(CanBusInterface):
    """Socket can driver interface, only available on Linux."""

    def __init__(self, channel = 'can0', auto_connect = True, *args, **kwargs):
        """Init candle driver interface
        
        Args:
            channel: socket can channel name
            auto_connect: if set to False, you need to call connect() manually
        """
        self._channel = channel
        self._bus = None
        super().__init__(auto_connect)
    
    def _enable(self):
        """Connect to socket can bus."""
        try:
            self._bus = can.interface.Bus(channel=self._channel, bustype='socketcan')
        except:
            raise IOError('Failed to open can bus.')
    
    def _disable(self):
        """Disconnect to socket can bus."""
        if self._bus is not None:
            self._bus.shutdown()
        del self._bus
    
    def _send(self, can_frame_id, data, remote):
        """Send can frame from candle interface."""
        msg = can.Message(arbitration_id=can_frame_id, data=data, is_extended_id=False, is_remote_frame=remote)
        try:
            self._bus.send(msg)
        except:
            raise IOError('Failed to send to can bus.')

    def _read(self, timeout):
        """Read can frame from candle interface."""
        try:
            msg = self._bus.recv(timeout)
            return msg.arbitration_id, msg.data, msg.timestamp
        except Exception as e:
            raise IOError('Failed to read from candle can bus.')



#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""monitor.py
Time    :   2022/10/17
Author  :   song 
Version :   1.0
Contact :   zhaosongy@126.com
License :   (C)Copyright 2022, robottime / robodyno

Monitor messages on can bus.
"""

import threading
from termcolor import colored

rx_lock = threading.Lock()

def rx_callback(device_id, command_id, data, timestamp):
    """Thread safe can bus rx callback, print message to console.
    
    Args:
        device_id: robodyno device id
        command_id: message command id
        data: message payload
        timestamp: time when message received
    """
    with rx_lock:
        payload = ''
        for byte in data:
            payload += '0x{:02x} '.format(byte)
        if len(payload) > 0:
            payload = payload[:-1]
        print('[{time}] {dev_id} | {cmd_id} [{payload}]'.format(
            time = '{}.{}'.format(int(timestamp/1000000), int(timestamp%1000000/1000)),
            dev_id = colored('0x{:02X}'.format(device_id), 'cyan'),
            cmd_id = colored('0x{:02X}'.format(command_id), 'magenta'),
            payload = colored(payload, 'yellow'),
        ))

def monitor(can, id_list):
    """Display messages from can bus to console.
    
    Args:
        can: can bus interface
        id_list: selected id list
    """
    if len(id_list) > 0:
        for id in id_list:
            can.subscribe(id, can.COMMAND_ID_ALL, rx_callback)
    else:
        can.subscribe(can.DEVICE_ID_ALL, can.COMMAND_ID_ALL, rx_callback)
    
    try:
        while True:
            pass
    except KeyboardInterrupt:
        if len(id_list) > 0:
            for id in id_list:
                can.unsubscribe(id, can.COMMAND_ID_ALL, rx_callback)
        else:
            can.unsubscribe(can.DEVICE_ID_ALL, can.COMMAND_ID_ALL, rx_callback)
        can.disconnect()



from setuptools import setup

long_description = open('README.md').read()

setup(
    name='robodyno',
    version='1.0',
    maintainer='robottime',
    maintainer_email='lab@robottime.cn',
    author='song',
    author_email='zhaosongy@126.com',
    description='The Robodyno Robot SDK for Python 3',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/robottime/Robodyno-Python-API',
    keywords=['robodyno', 'robot', 'robot module'],
    license='MIT License',
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX :: Linux',
        'Framework :: Robot Framework',
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: Science/Research',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
    ],
    packages=[
        'robodyno', 
        'robodyno.components', 
        'robodyno.interfaces', 
        'robodyno.tools',
        'robodyno.interfaces.can_bus',
        'robodyno.interfaces.webots',
        'robodyno.robots',
        'robodyno.robots.six_dof_collaborative_robot',
        'robodyno.robots.utils',
    ],
    package_dir={'robodyno.robots': 'robodyno_robots'},
    python_requires='>=3.6',
    install_requires=['numpy>=1.10.0'],
    extras_require={
        ':sys_platform == "win32"': [
            'candle_driver'
        ],
        ':"linux" in sys_platform': [
            'python-can>=3.2.0, <4.0'
        ],
    },
    entry_points={
        'console_scripts': [
            'robodyno = robodyno:robodyno',
            'robodyno-motor = robodyno:robodyno_motor'
        ]
    }
)



import sys
import time
import argparse
import colorama
from termcolor import colored
sys.path.append("..")
from robodyno_components_python.interfaces import CanBus
from robodyno_components_python.components import Motor

colorama.init(autoreset=True)

parser = argparse.ArgumentParser(description='Robodyno command line tools.')
parser.add_argument('command', 
                    choices=['enable', 'disable', 'info', 'control', 'config', 'reset'], 
                    help='Operate robodyno motor via command line.')
parser.add_argument('--id', type=str, nargs='*', default=['0x10'],
                    help='robodyno device id filter, dafault [0x10]')
parser.add_argument('-c', '--channel', type=str, default='can0',
                    help='can bus channel')
parser.add_argument('-b', '--bitrate', type=str, default='CAN_1M',
                    help='can bus bitrate')

parser.add_argument('-p', '--pos', type=float,
                    help='motor position setpoint(rad)')
parser.add_argument('-v', '--vel', type=float,
                    help='motor velocity setpoint(rad/s)')
parser.add_argument('-t', '--torque', type=float,
                    help='motor torque setpoint(Nm)')
parser.add_argument('--new-id', type=str,
                    help='change motor id with new id')
parser.add_argument('-s', '--save', action='store_true',
                    help='save motor configuration when set new id')

args = parser.parse_args()

try:
    can = CanBus(bitrate = args.bitrate, channel = args.channel)
except:
    raise ValueError('Wrong can bus configurations.')

def id_to_motor_list(str_list):
    """Parse id list to motor object list.
    
    Args:
        str_list: string id list
    
    Returns:
        robodyno motor object list
    """
    id_list = []
    motor_list = []
    if len(str_list) > 0 and ('all' in str_list):
        id_list = [id for id in range(0x01, 0x40)]
    else:
        id_list = list(map(lambda id_str: int(id_str, 0), str_list))
    for id in id_list:
        try:
            motor_list.append(Motor(can, id))
        except:
            pass
    return motor_list

try:
    motor_list = id_to_motor_list(args.id)
except:
    raise ValueError('Wrong id list.')

if len(motor_list) == 0:
    raise ValueError('No available motor on can bus.')

if args.command == 'enable':
    for motor in motor_list:
        motor.enable()

if args.command == 'disable':
    for motor in motor_list:
        motor.disable()

if args.command == 'info':
    for motor in motor_list:
        time.sleep(1)
        ver = motor.get_version(0.1)
        vbus = motor.get_voltage(0.1)
        temp = motor.get_temperature(0.1)
        pos, vel, torque = motor.get_feedback(0.1)
        state_str = 'mode: ' + colored(motor.mode.name, 'magenta')
        state_str += ', state: '
        if motor.error['error'].value:
            state_str += colored(motor.error['error'].name, 'red')
        else:
            state_str += colored(motor.state.name, 'blue')
        print(colored('[0x{:02X}]'.format(motor.id), 'cyan') + ' ' +
                colored(ver['type'].name, 'green') + ' ' +
                'ver: {}.{}'.format(ver['main_version'], ver['sub_version']))
        print(state_str)
        print('vbus: {:.2f}V, temp: {:.2f}°C'.format(vbus, temp))
        print('pos: {:.4f}, vel: {:.4f}, torque: {:.4f}'.format(pos, vel, torque))

if args.command == 'control':
    for motor in motor_list:
        if args.pos is not None:
            vel_ff = args.vel if args.vel is not None else 0
            torque_ff = args.torque if args.torque is not None else 0
            for motor in motor_list:
                motor.position_filter_mode(5)
                motor.enable()
                motor.set_pos(args.pos, vel_ff, torque_ff)
        elif args.vel is not None:
            torque_ff = args.torque if args.torque is not None else 0
            for motor in motor_list:
                motor.velocity_ramp_mode(5)
                motor.enable()
                motor.set_vel(args.vel, torque_ff)
        elif args.torque is not None:
            for motor in motor_list:
                motor.torque_mode(5)
                motor.enable()
                motor.set_torque(args.torque)

if args.command == 'config':
    if len(motor_list) != 1:
        raise ValueError('Can not set multiple motor ids at the same time.')
    motor = motor_list[0]
    try:
        new_id = int(args.new_id, 0)
        if new_id < 0x01 or new_id >= 0x40:
            raise ValueError()
    except:
        raise ValueError('Wrong motor new id, choose from [0x01,0x40).')
    motor.config_can_bus(new_id, bitrate = args.bitrate)
    time.sleep(0.5)
    motor = Motor(can, new_id)
    if args.save:
        motor.save_configuration()

if args.command == 'reset':
    if len(motor_list) != 1:
        raise ValueError('Can not reset multiple motor at the same time.')
    motor = motor_list[0]
    print('resetting...')
    motor.reset()
    time.sleep(3)
    motor = Motor(can, 0x10)
    print('calibrating...')
    motor.calibrate()
    time.sleep(1)
    while motor.state.value != 1:
        pass
    time.sleep(0.2)
    motor.save_configuration()
    time.sleep(0.5)
    print('finished.')



#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""interface.py
Time    :   2022/09/30
Author  :   song 
Version :   1.0
Contact :   zhaosongy@126.com
License :   (C)Copyright 2022, robottime / robodyno

Robodyno Interface Abstract Class
"""

from abc import ABC, abstractmethod

class Interface(ABC):
    """Interface with get node methods."""

    @abstractmethod
    def get_device_node(self, device):
        """Get common robodyno device node from interface.
        
        Args:
            device: robodyno device object
        
        Returns:
            device node with specific interface
        """
    
    @abstractmethod
    def get_motor_node(self, motor):
        """Get motor node from current interface.
        
        Args:
            motor: robodyno motor object
        
        Returns:
            motor node with specific interface
        """
